import redis
import websockets
import json

from abc import ABC, abstractmethod
from dvenv import log
from dvenv.transact import REQUEST_QUEUE_TOPIC
from asgiref.sync import async_to_sync


class Broker(ABC):
    def __init__(self, *, host, port, request_timeout, response_timeout) -> None:
        super().__init__()
        self.host = host
        self.port = port
        self.request_timeout = request_timeout
        self.response_timeout = response_timeout

    @abstractmethod
    def send_request(self, request):
        pass

    @abstractmethod
    def send_message(self, request, topic):
        pass

    @abstractmethod
    def send_response(self, response, topic):
        pass

    @abstractmethod
    def is_connected(self):
        pass

    @abstractmethod
    def subscribe(self, topic):
        pass

    @abstractmethod
    def get_messages(self):
        pass


class RedisBroker(Broker):
    def __init__(self, *, host, port, request_timeout, response_timeout) -> None:
        super().__init__(
            host=host,
            port=port,
            request_timeout=request_timeout,
            response_timeout=response_timeout,
        )

        self.broker_backend = "redis"

        log.action("********************************")
        log.action(f"BROKER: {self.broker_backend}")
        log.action(f"HOST: {self.host}")
        log.action(f"PORT: {self.port}")
        log.action(f"REQUEST TIMEOUT: {self.request_timeout}")
        log.action(f"RESPONSE TIMEOUT: {self.response_timeout}")
        log.action("********************************")

        self._connect()

    def _connect(self):
        self.r = redis.StrictRedis(host=self.host, port=self.port, db=0)

        if not self.is_connected():
            log.die(f"Unable to connect to broker: redis@{self.host}/{self.port}")

        log.info(f"Connected to broker: redis@{self.host}/{self.port}")
        self.ps = self.r.pubsub()

    def is_connected(self):
        try:
            self.r.ping()
        except redis.exceptions.ConnectionError:
            return False

        return True

    def send_request(self, request):
        payload = request.serialize()
        self.r.publish(REQUEST_QUEUE_TOPIC, payload)

    def send_message(self, request, topic):
        payload = request.serialize()
        self.r.publish(topic, payload)

    def send_response(self, response, topic):
        payload = response.serialize()
        self.r.publish(topic, payload)

    def subscribe(self, topic):
        return self.ps.subscribe(topic)

    def get_messages(self):
        messages = []

        message = self.ps.get_message()
        while message is not None:
            messages.append(message)
            message = self.ps.get_message()

        return messages



import click
import sys

from dvenv.commands import host
from dvenv.commands import init


@click.group()
def entry_point():
    pass


def main():
    entry_point.add_command(host.host)
    entry_point.add_command(init.init)
    entry_point()


if __name__ == "__main__":
    sys.exit(main())



import click
import sys
import os

OUTPUT_PREFIX = "dvenv: "


def warn(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='red')}"  # noqa
    )


def debug(str):
    if int(os.getenv("DVENV_DEBUG", 0)) != 0:
        click.echo(
            f"DEBUG: {click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='white')}"  # noqa
        )


def die(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}FATAL: {click.style(str, bg='black', fg='red')}"  # noqa
    )
    sys.exit(1)


def info(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='white')}"  # noqa
    )


def action(str, prepend=""):
    click.echo(
        f"{prepend}{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='yellow')}"  # noqa
    )


def warn_confirm(str):
    return click.confirm(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='red')}",  # noqa
    )


def action_confirm(_str):
    return click.confirm(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(_str, bg='black', fg='yellow')}",  # noqa
    )


def action_prompt(str, type, default=""):
    return click.prompt(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='yellow')}",  # noqa
        type=type,
        show_default=True,
        default=default,
    )



import click

from dvenv import log
from dvenv.server import Server
from dvenv.decorators import with_config


@click.command()
@with_config
def host(cfg):
    """Host a dvenv environment for clients."""

    server = Server(cfg=cfg)
    server.run()



import time
import json

from dvenv.broker import RedisBroker
from dvenv.transact import Response, Commands, RequestError, REQUEST_QUEUE_TOPIC
from dvenv.channel import ChannelManager
from dvenv import log


BROKER_SLEEP_TIME = 0.001


class Server:
    def __init__(self, *, cfg) -> None:
        log.action("Starting dvenv server")

        self.broker = RedisBroker(
            host=cfg["broker_host"],
            port=cfg["broker_port"],
            request_timeout=cfg["request_timeout"],
            response_timeout=cfg["response_timeout"],
        )

        self.environment_path = cfg["environment_path"]
        self.channels = {}
        self.environments = 0
        self.max_channels = 10  # TODO: add to dvenv config file
        self.max_environments = cfg["max_environments"]

    def run(self):
        self.broker.subscribe(REQUEST_QUEUE_TOPIC)

        while self.broker.is_connected():
            messages = self.broker.get_messages()
            for message in messages:
                log.debug(f"Received message: {message}")
                self._handle_message(message)

            for channel_id in self.channels.keys():
                self.channels[channel_id].process()

            time.sleep(BROKER_SLEEP_TIME)

    def _handle_message(self, message):
        channel_id = message["channel"].decode()
        if channel_id == REQUEST_QUEUE_TOPIC:
            try:
                self._handle_request(message["data"].decode())
            except AttributeError:
                pass
        elif channel_id in self.channels.keys():
            try:
                self._handle_channel_message(channel_id, message["data"].decode())
            except AttributeError:
                pass
        else:
            log.warn(f"Unhandled message: {message}")

    def _handle_request(self, request_data):
        request_data = json.loads(request_data)
        cmd_id = request_data["cmd_id"]
        channel_id = str(request_data["data"]["channel_id"])

        if cmd_id == Commands.INIT:
            log.action("Received INIT request")

            if len(self.channels.keys()) >= self.max_channels:
                log.warn("Rejected INIT request: No availability")

                response = Response(
                    cmd_id=Commands.INIT, data={"error": RequestError.NO_AVAILABILITY}
                )
                return self.broker.send_response(response, topic=channel_id)

            else:
                log.action("Accepted INIT request")
                response = Response(cmd_id=Commands.INIT, data={})

                if channel_id not in self.channels.keys():
                    log.action(f"Opening new channel: {channel_id}")
                    self.channels[channel_id] = ChannelManager(
                        channel_id=channel_id, broker=self.broker
                    )
                    self.broker.subscribe(channel_id)

                return self.broker.send_response(response, topic=channel_id)

    def _handle_channel_message(self, channel_id, msg):
        msg = json.loads(msg)
        msg_data = msg.get("data")

        client_id = msg_data.get("client_id")
        if client_id is None:
            return

        self.channels[channel_id].on_message(msg)



import click

from dvenv import log
from dvenv.decorators import with_client


@click.command()
@with_client
def init(client):
    """Initialize a new dvenv environment."""

    return_code = client.create_new_environment(
        python_version="python3.8", path="/tmp/test2"
    )

    if return_code == 0:
        log.action("Created new virtual environment")
    else:
        log.die(f"Could not create a new environment, return code: {return_code}")



import asyncio
import sys
import select
import threading
import time

from enum import Enum
from queue import Queue
from dvenv import log
from dvenv.transact import Commands, Response
from collections import namedtuple

Command = namedtuple("Command", "id cmd")


class ChannelCommand(Enum):
    PYTHON_INTERACTIVE = Command(id=0, cmd=["python", "-i"])
    PYTHON = Command(id=1, cmd=["python"])
    PIP = Command(id=2, cmd=["pip"])
    CREATE_ENV = Command(id=3, cmd=["virtualenv", "-p"])


BUFFER_ORDER = ["stdout", "stderr"]
CARRIAGE_RETURN = "\r\n"
TIMEOUT = 0.01


class ChannelManager:
    def __init__(self, *, channel_id, broker) -> None:
        self.channel_id = channel_id
        self.broker = broker
        self.worker = ChannelWorker()
        self.running_process = False

        log.action(f"Initialized channel manager on: {self.channel_id}")

    def on_message(self, msg):
        cmd_id = msg["cmd_id"]

        if cmd_id == Commands.CREATE_ENV:
            log.action("Received CREATE_ENV command")

            python_version = msg["data"]["python_version"]
            path = msg["data"]["path"]

            self.running_process = True
            self.worker.run(ChannelCommand.CREATE_ENV, python_version, path)

            # if self.environments >= self.max_environments:
            #     log.warn("No availability, rejecting CREATE_ENV")

            #     response = Response(
            #         cmd_id=Commands.INIT, data={"error": RequestError.NO_AVAILABILITY}
            #     )

            #     return self.broker.send_response(response, topic=self.channel_id)

        elif cmd_id == Commands.RUN:
            log.action("Received RUN command")

            channel_command = msg["data"]["channel_command"]
            args = msg["data"]["args"]

            if self.running_process and self.worker.active_process:
                self.worker.active_process.terminate()

            if channel_command == list(ChannelCommand.PYTHON_INTERACTIVE.value):
                self.worker.run(ChannelCommand.PYTHON_INTERACTIVE, *args)
                self.running_process = True
            elif channel_command == list(ChannelCommand.PYTHON.value):
                self.worker.run(ChannelCommand.PYTHON, *args)
                self.running_process = True
            else:
                pass

        elif cmd_id == Commands.PROCESS_INPUT:
            self.worker.send_line(line=msg["data"]["input"])

    def kill(self):
        if self.worker.active_process:
            self.worker.active_process.terminate()
            self.worker.reset()
            self.running_process = False

    def status(self):
        pass

    def process(self):
        reset = False
        return_code = None

        if self.worker.active_process:
            process_status = self.worker.process_status()

            if process_status is not None:
                return_code = process_status
                reset = True

            output = self.worker.get_output()

            if output != "":
                response = Response(
                    cmd_id=Commands.PROCESS_OUTPUT,
                    data={"output": output},
                )

                self.broker.send_response(response, topic=self.channel_id)

        if reset:
            response = Response(
                cmd_id=Commands.PROCESS_EXIT,
                data={"return_code": return_code},
            )

            self.broker.send_response(response, topic=self.channel_id)
            self.worker.reset()


class ChannelWorker:
    def __init__(self) -> None:
        self.reset()

    async def _read_stream(self, proc, stream, out_queue):
        timeout = TIMEOUT

        while True:
            try:
                c = await asyncio.wait_for(stream.read(1), timeout)
            except asyncio.TimeoutError:
                c = None

            if c:
                out_queue.put(c.decode())

            if proc.returncode is not None:
                break

    def _write_stream(self, proc, stream):
        timeout = TIMEOUT

        while proc.returncode is None:
            input, _, _ = select.select([sys.stdin], [], [], timeout)

            if input:
                line = sys.stdin.readline().rstrip() + CARRIAGE_RETURN
                stream.write(line.encode("utf-8"))

    async def _async_callback(self, cmd, proc_queue, stdout_queue, stderr_queue, *args):
        extra_args = list(args[0])

        proc = await asyncio.create_subprocess_exec(
            *cmd.value.cmd,
            *extra_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
        )

        proc_queue.put(proc)

        await asyncio.wait(
            [
                self._read_stream(proc, proc.stdout, stdout_queue),
                self._read_stream(proc, proc.stderr, stderr_queue),
            ]
        )

        return await proc.wait()

    def _flush_queue(self, queue):
        data = []

        while not queue.empty():
            c = queue.get()
            data.append(c)

        return data

    def _spawn_async(self, cmd, proc_queue, stdout_queue, stderr_queue, *args):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(
            self._async_callback(cmd, proc_queue, stdout_queue, stderr_queue, args)
        )
        loop.close()

    def run(self, cmd, *args):
        log.action(f"Running: {cmd}")

        self.proc_queue = Queue()
        self.stdout_queue = Queue()
        self.stderr_queue = Queue()

        self.read_thread = threading.Thread(
            target=self._spawn_async,
            args=(cmd, self.proc_queue, self.stdout_queue, self.stderr_queue, *args),
        )
        self.read_thread.daemon = True
        self.read_thread.start()

        self.active_process = self.proc_queue.get()

    def reset(self):
        self.read_thread = None
        self.active_process = None
        self.proc_queue = None
        self.stdout_queue = None
        self.stderr_queue = None

    def send_line(self, *, line):
        if self.active_process is not None:
            line = line.rstrip() + CARRIAGE_RETURN
            self.active_process.stdin.write(line.encode("utf-8"))

    def process(self):
        data = []

        if BUFFER_ORDER[0] == "stderr":
            data = self._flush_queue(self.stderr_queue)
            data.extend(self._flush_queue(self.stdout_queue))
        elif BUFFER_ORDER[0] == "stdout":
            data = self._flush_queue(self.stdout_queue)
            data.extend(self._flush_queue(self.stderr_queue))

        return data

    def get_output(self):
        data = self.process()
        output_str = "".join(data)
        return output_str

    def wait_for_completion(self):
        if self.active_process is not None:

            self.read_thread.join()
            return_code = self.active_process.returncode
            self.reset()

            return return_code

    def process_status(self):
        if self.active_process is not None:
            if self.active_process.returncode is not None:
                return self.active_process.returncode



from functools import wraps
from dvenv import log
from dvenv.client import Client
from dvenv.config import load_configuration
from dvenv.exceptions import NoConfigurationFound, InvalidConfiguration


def with_client(callback):
    @wraps(callback)
    def wrapper(*args, **kwargs):

        try:
            cfg = load_configuration()
        except NoConfigurationFound:
            log.die("No configuration found.")
        except InvalidConfiguration:
            log.die("Invalid configuration.")

        client = Client(cfg=cfg)
        client.initialize_channel()

        return callback(client, *args, **kwargs)

    return wrapper


def with_config(callback):
    @wraps(callback)
    def wrapper(*args, **kwargs):

        try:
            cfg = load_configuration()
        except NoConfigurationFound:
            log.die("No configuration found.")
        except InvalidConfiguration:
            log.die("Invalid configuration.")

        return callback(cfg, *args, **kwargs)

    return wrapper



import json

from enum import IntEnum
from uuid import uuid4

REQUEST_QUEUE_TOPIC = "request_queue"


class Commands(IntEnum):
    INIT = 0
    CREATE_ENV = 1
    DELETE = 2
    RUN = 3
    PROCESS_INPUT = 4
    PROCESS_OUTPUT = 5
    PROCESS_EXIT = 6


class RequestResponseType(IntEnum):
    SYNC = 0
    ASYNC = 1


class RequestError(IntEnum):
    NO_AVAILABILITY = 1


class Request:
    def __init__(self, *, cmd_id, data) -> None:
        self.id = str(uuid4())[:8]
        self.cmd_id = cmd_id
        self.data = data

    def serialize(self):
        payload = {"id": self.id, "cmd_id": self.cmd_id, "data": self.data}
        return json.dumps(payload)


class Response:
    def __init__(self, *, cmd_id, data) -> None:
        self.id = str(uuid4())[:8]
        self.cmd_id = cmd_id
        self.data = data

    def serialize(self):
        payload = {"id": self.id, "cmd_id": self.cmd_id, "data": self.data}
        return json.dumps(payload)



import os
import time
import json
import sys
import select

from uuid import uuid4
from pathlib import Path
from dvenv.broker import RedisBroker
from dvenv.transact import Request, Commands, RequestError
from dvenv.channel import ChannelCommand
from dvenv import log

DVENV_CHANNEL_PATH = "./.dvenv/channel"
CARRIAGE_RETURN = "\r\n"
TIMEOUT = 0.001


class Client:
    def __init__(self, *, cfg) -> None:
        self.cfg = cfg
        self.broker = RedisBroker(
            host=cfg["broker_host"],
            port=cfg["broker_port"],
            request_timeout=cfg["request_timeout"],
            response_timeout=cfg["response_timeout"],
        )

        self.environment_path = cfg["environment_path"]
        self.client_id = str(uuid4())
        self.channel_id = self._get_channel_id()
        self.broker.subscribe(self.channel_id)

        self.token = "user1"
        self.msg_backlog = []

    def _get_channel_id(self):
        if os.path.exists(DVENV_CHANNEL_PATH):
            with open(DVENV_CHANNEL_PATH, "r") as f:
                return f.read()
        else:
            return str(uuid4())

    def initialize_channel(self):
        req = Request(
            cmd_id=Commands.INIT,
            data={
                "token": self.token,
                "client_id": self.client_id,
                "channel_id": self.channel_id,
            },
        )
        self.broker.send_request(req)

        response = self._wait_for_response(
            cmd_id=Commands.INIT, timeout=self.cfg["response_timeout"]
        )

        if response is None:
            log.die("No response received from broker")

        response_data = response.get("data")
        if response_data == {}:
            log.action("INIT request accepted")
        elif response_data.get("error") == RequestError.NO_AVAILABILITY:
            log.die("INIT request rejected: server is at capacity")

        Path(".dvenv").mkdir(parents=True, exist_ok=True)
        with open(DVENV_CHANNEL_PATH, "w") as f:
            f.write(self.channel_id)

    def create_new_environment(
        self,
        *,
        python_version,
        path,
    ):
        req = Request(
            cmd_id=Commands.CREATE_ENV,
            data={
                "token": self.token,
                "client_id": self.client_id,
                "channel_id": self.channel_id,
                "python_version": python_version,
                "path": path,
            },
        )
        self.broker.send_message(req, topic=self.channel_id)
        return self._enter_interactive_process()

    def run_python(self, *args):
        req = None

        if len(args) == 0:
            req = Request(
                cmd_id=Commands.RUN,
                data={
                    "channel_command": ChannelCommand.PYTHON_INTERACTIVE.value,
                    "args": list(args),
                    "token": self.token,
                    "client_id": self.client_id,
                    "channel_id": self.channel_id,
                },
            )
        else:
            req = Request(
                cmd_id=Commands.RUN,
                data={
                    "channel_command": ChannelCommand.PYTHON.value,
                    "args": list(args),
                    "token": self.token,
                    "client_id": self.client_id,
                    "channel_id": self.channel_id,
                },
            )

        self.broker.send_message(req, topic=self.channel_id)
        return self._enter_interactive_process()

    def receive_messages(self):
        response_messages = self.broker.get_messages()

        if len(response_messages) > 0:
            for message in response_messages:
                try:
                    _ = self._parse_channel_message(message["data"].decode())
                except AttributeError:
                    pass

                response_messages = self.broker.get_messages()

    def _enter_interactive_process(self):
        process_running = True
        return_code = None

        while process_running:
            self._get_input()

            response_messages = self.broker.get_messages()
            if len(response_messages) > 0:
                for message in response_messages:
                    try:
                        msg = self._parse_channel_message(message["data"].decode())

                        cmd_id = msg.get("cmd_id")
                        if cmd_id is not None and cmd_id == Commands.PROCESS_OUTPUT:
                            self._flush_output(msg)
                        elif cmd_id is not None and cmd_id == Commands.PROCESS_EXIT:
                            return_code = msg["data"]["return_code"]
                            log.action(
                                f"Process exited with return code: {return_code}",
                                prepend="\n",
                            )
                            process_running = False
                        elif cmd_id is not None and cmd_id == Commands.CREATE_ENV:
                            pass
                        else:
                            self.msg_backlog.append(self.msg_backlog)
                    except AttributeError:
                        pass

        return return_code

    def _get_input(self):
        input, _, _ = select.select([sys.stdin], [], [], TIMEOUT)

        if input:
            line = sys.stdin.readline().rstrip() + CARRIAGE_RETURN

            req = Request(
                cmd_id=Commands.PROCESS_INPUT,
                data={
                    "token": self.token,
                    "client_id": self.client_id,
                    "channel_id": self.channel_id,
                    "input": line,
                },
            )
            self.broker.send_message(req, topic=self.channel_id)

    def _flush_output(self, msg):
        sys.stdout.write(msg["data"]["output"])
        sys.stdout.flush()

    def _wait_for_response(self, *, cmd_id, timeout):
        start_time = time.time()
        elapsed = 0.0

        while elapsed < timeout:
            response_messages = self.broker.get_messages()

            if len(response_messages) > 0:
                for message in response_messages:
                    try:
                        msg = self._parse_channel_message(message["data"].decode())
                        cmd_id = msg.get("cmd_id")
                        if cmd_id is not None and cmd_id == cmd_id:
                            return msg
                        else:
                            self.msg_backlog.append(self.msg_backlog)
                    except AttributeError:
                        pass

            elapsed = time.time() - start_time

    def _parse_channel_message(self, msg):
        msg = json.loads(msg)
        msg_data = msg["data"]

        client_id = msg_data.get("client_id")
        if client_id is not None:
            return

        return msg



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dvenv', 'dvenv.commands']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'asgiref>=3.4.1,<4.0.0',
 'click>=8.0.1,<9.0.0',
 'redis>=3.5.3,<4.0.0']

entry_points = \
{'console_scripts': ['dpython = dvenv:runner.python',
                     'dvenv = dvenv:main.main']}

setup_kwargs = {
    'name': 'dvenv',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Luke Lombardi',
    'author_email': 'luke@slai.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import yaml
import os

from dvenv.exceptions import NoConfigurationFound, InvalidConfiguration


def _validate_configuration(cfg):
    return True


def load_configuration(path="dvenv.yml"):
    if not os.path.exists(path):
        raise NoConfigurationFound

    with open(path, "r") as f:
        try:
            cfg = yaml.safe_load(f.read())
        except yaml.YAMLError:
            raise InvalidConfiguration

    is_valid_config = _validate_configuration(cfg)
    if not is_valid_config:
        raise InvalidConfiguration

    return cfg
